from charms.reactive import when, when_not, set_state, remove_state, hook
from charms.layer.ruby import bundle
from charmhelpers.core.templating import render
from charmhelpers.core.host import service_running, service
from charmhelpers.core import hookenv
from charmhelpers.core.hookenv import open_port
import psutil
import os.path
import os
import pwd
import grp
import time
import subprocess


# Utility function to read the discourse.conf file, which is just key=value
# and turn it into a dictionary that we return so it can be added to.
def ingest_config():
    discourse_conf = {}
    if os.path.isfile('/srv/discourse/current/config/discourse.conf'):
        with open('/srv/discourse/current/config/discourse.conf') as config:
            for line in config:
                (key, value) = line.split('=')
                discourse_conf[key.strip()] = value.strip()
    return discourse_conf


# Utility function to write a dictionary as key=value pairs to the config
# location
def write_config(config):
    render(source="discourse.conf",
           target="/srv/discourse/current/config/discourse.conf",
           owner="www-data",
           perms=0o644,
           context={'config': config})


# Ingest the existing config, so we don't lose ad hoc edits or initial config
# then add database configuration to it and write it out.
def write_db_config(pgsql):
    config = ingest_config()
    db_config = pgsql.master
    config['db_name'] = db_config.dbname
    config['db_host'] = db_config.host
    config['db_username'] = db_config.user
    config['db_password'] = db_config.password
    write_config(config)
    set_state('discourse.database.configured')


# Utility function to fetch plugins
def fetch_plugins():
    # Our config option should be a comma separated list of github URLs
    for plugin_url in hookenv.config('plugins').split(','):
        # This splits the /'s the URL and gives us the final directory part
        plugin_dir = plugin_url.split('/')[-1]
        # Then split the .git off if it exists
        plugin_dir = plugin_dir.split('.')[0]
        # Then add our static path to it, should reference the app-path config
        plugin_dir = "/srv/discourse/current/plugins/" + plugin_dir
        git_proxy = ''
        if hookenv.config('http-proxy'):
            git_proxy = '-c "http.proxy={}"'.format(hookenv.config('http-proxy'))
        if os.path.isdir(plugin_dir):
            git_command = '-C "{}" pull'.format(plugin_dir)
        else:
            git_command = '-C "/srv/discourse/current/plugins/" clone'
        clone_cmd = "/usr/bin/git {} {} {}".format(git_proxy, git_command, plugin_url)
        subprocess.check_call(clone_cmd, shell=True)


# When we connect to the database units, request our database and extensions
@when_not('discourse.database.created')
@when('db.connected')
def request_db(pgsql):
    pgsql.set_database('discourse')
    pgsql.set_extensions(['hstore', 'pg_trgm'])
    set_state('discourse.database.created')


# I'm not 100% sure how this works?  But it does.
@hook('install')
def git_proxy():
    proxy = hookenv.config('http-proxy')
    if proxy:
        os.environ['HTTP_PROXY'] = proxy
        os.environ['HTTPS_PROXY'] = proxy
        os.environ['http_proxy'] = proxy
        os.environ['https_proxy'] = proxy


# The git-deploy layer exposes codebase.available when it's been successfully
# cloned to the location.  Check to make sure ruby is available as well, then
# write our base config, set the status and install all the gems we need.
@when('codebase.available')
@when('ruby.available')
@when_not('discourse.installed')
def install_discourse():
    config = {'hostname': hookenv.config('hostname')}
    config['smtp_address'] = "127.0.0.1"
    # Required to work with a stock install postfix
    config['smtp_openssl_verify_mode'] = "none"
    config['smtp_domain'] = hookenv.config('hostname')
    # There's no config default for this, so only write it if it's set
    if hookenv.config('admin-users'):
        config['developer_emails'] = hookenv.config('admin-users')
    write_config(config)
    # Bionic seems to require that you manually install rake globally before
    # running bundle install, as far as I can tell this shouldn't hurt anything
    # in the xenial version, it's just not necessary per se there.
    rake_install_cmd = ['gem', 'install', 'rake']
    proxy = hookenv.config('http-proxy')
    if proxy:
        rake_install_cmd.extend(['-p', 'http://squid.internal:3128'])
    subprocess.call(rake_install_cmd)
    # Bundle command is from the ruby layer, it'll install our gem dependencies
    bundle('install')
    hookenv.status_set('blocked', 'Discourse installed, waiting on database.')
    set_state('discourse.installed')


# If we haven't already written the database config (the state is set in the
# utility function) and the codebase is installed and ready and we have a
# connected DB master, write the DB config to discourse.conf
@when_not('discourse.database.configured')
@when('discourse.installed')
@when('db.master.available')
def db_available(pgsql):
    write_db_config(pgsql)


# Discourse requires some prep, so once we have a database and have written the
# config, run the various setup tasks.  Some of this needs to be done in other
# circumstances, so we may want to genericize the rakes out into a function.
@when_not('discourse.codebase.prepared')
@when('db.master.available')
@when('discourse.database.configured')
@when('discourse.database.created')
def prepare_codebase(pgsql):
    # Create all the Discourse specific DB schema
    bundle('exec rake db:migrate RAILS_ENV=production')
    # Compile CSS/JS?  I'm not sure what this does but it's necessary.
    bundle('exec rake assets:precompile RAILS_ENV=production')
    if hookenv.config('plugins'):
        fetch_plugins()
    # The ruby layer doesn't have a concept of ownership, so after the rakes
    # have run, hit everything with a chown hammer so the app server doesn't choke
    subprocess.call(["/bin/chown", "-R", "www-data:www-data", "/srv/discourse/current/"])
    # Set some states to trigger server configuration off of
    set_state('discourse.configure.unicorn')
    set_state('discourse.configure.sidekiq')
    set_state('discourse.codebase.prepared')


# Triggered by the state set above, make a required path if it doesn't exist
# then write out the sytemd service file and start the server.
@when('discourse.configure.unicorn')
def configure_unicorn():
    if not os.path.isdir('/srv/discourse/current/tmp/pids'):
        os.makedirs('/srv/discourse/current/tmp/pids')
    # There's a symlink owned by root that makes unicorn choke consistently so
    # let's chown it to the running user.  No, you shouldn't run and own files
    # as the same user but unicorn/rails seem to require it. ¯\_(ツ)_/¯
    uid = pwd.getpwnam('www-data').pw_uid
    gid = grp.getgrnam('www-data').gr_gid
    os.chown('/srv/discourse/current/tmp/pids', uid, gid)
    # This is a symlink that is root owned and unicorn horks if it isn't owned
    # by the app server running user.
    os.chown('/srv/discourse/current/public/plugins/discourse-narrative-bot', uid, gid)
    render(source="unicorn.service",
           target="/lib/systemd/system/unicorn.service",
           perms=0o644,
           context={'config': hookenv.config})
    service('enable', 'unicorn')
    if service_running('unicorn'):
        service('restart', 'unicorn')
    else:
        service('start', 'unicorn')
    # We don't technically need to open the port, but there's no harm and it's
    # not a bad idea.
    open_port(3000, protocol='tcp')
    # Set the state to active, so we have prettier statu output.
    hookenv.status_set('active', 'Discourse running.')
    remove_state('discourse.configure.unicorn')


# Largely the same as the unicorn setup but with less weird chowning required.
@when('discourse.configure.sidekiq')
def configure_sidekiq():
    render(source="sidekiq.service",
           target="/lib/systemd/system/sidekiq.service",
           perms=0o644,
           context={'config': hookenv.config})
    service('enable', 'sidekiq')
    if service_running('sidekiq'):
        service('restart', 'sidekiq')
    else:
        service('start', 'sidekiq')
    remove_state('discourse.configure.sidekiq')


# State provided by the website interface layer, it handles setting up the
# provider side of the website relation with whatever the frontend/balancer
# will be.
@when('website.available')
def configure_website(website):
    hostname = hookenv.config('hostname')
    website.configure(port=3000, hostname=hostname)


# Set up Nagios checks when the nrpe-external-master subordinate is related
@when('nrpe-external-master.available')
def setup_nagios(nagios):
    config = hookenv.config()
    unit_name = hookenv.local_unit()
    check_base = '/usr/lib/nagios/plugins/'
    process_check = check_base + 'check_procs'

    unicorn_check = [process_check, '-c', '4:4', '-a', 'unicorn']
    nagios.add_check(unicorn_check, name="unicorn_process",
                     description="Check for Unicorn processes",
                     context=config['nagios_context'],
                     servicegroups=config['nagios_servicegroups'],
                     unit=unit_name)

    sidekiq_check = [process_check, '-c', '1:1', '-a', 'sidekiq']
    nagios.add_check(sidekiq_check, name="sidekiq_process",
                     description="Check for Sidekiq processes",
                     context=config['nagios_context'],
                     servicegroups=config['nagios_servicegroups'],
                     unit=unit_name)

    unicorn_http_check = [check_base + 'check_http', '-H', '127.0.0.1', '-p', '3000']
    nagios.add_check(unicorn_http_check, name="unicorn_port",
                     description="Check Unicorn app server port",
                     context=config['nagios_context'],
                     servicegroups=config['nagios_servicegroups'],
                     unit=unit_name)

    postfix_check = [check_base + 'check_tcp', '-H', '127.0.0.1', '-p', '25']
    nagios.add_check(postfix_check, name="postfix_port",
                     description="Check postfix port",
                     context=config['nagios_context'],
                     servicegroups=config['nagios_servicegroups'],
                     unit=unit_name)

    redis_check = [check_base + 'check_tcp', '-H', '127.0.0.1', '-p', '6379']
    nagios.add_check(redis_check, name="redis_port",
                     description="Check redis port",
                     context=config['nagios_context'],
                     servicegroups=config['nagios_servicegroups'],
                     unit=unit_name)


# It's possible to make this more granular by firing on particular config
# items changing, but it's not really necessary for us, a lot of this stuff
# is only really relevant on initial setup and can be done through the web
# UI once Discourse is up and running, so just rewrite the whole config when
# one of the items changes.
@when('config.changed')
def write_new_config():
    config = ingest_config()
    config['hostname'] = hookenv.config('hostname')
    config['smtp_domain'] = hookenv.config('hostname')
    admin_users = hookenv.config('admin-users')
    if admin_users:
        config['developer_emails'] = admin_users
    write_config(config)
    remove_state('discourse.codebase.prepared')
